import os
import sys
from loguru import logger
from .constants import *

# --------------------------------------------------------------------------------------------------------------
# (*) 로그 레벨 우선순위 (loguru 기본 레벨)
#   - TRACE     – 5  (가장 낮음, 가장 상세한 로그)
#   - DEBUG     – 10
#   - INFO      – 20
#   - SUCCESS   – 25 (특정 작업이 성공했을 때, 커스텀)
#   - WARNING   – 30
#   - ERROR     – 40
#   - CRITICAL  – 50 (가장 심각한 오류)
# --------------------------------------------------------------------------------------------------------------
# (*) logger.add() : backtrace, diagnose 옵션
#   - logger.exception() 에만 영향을 미침 (주의 logger.error(), critical() 에는 영향 안미침)
#   - backtrace, diagnose 모두 True 가 기본값 : 즉, 모든 스택이 나열되면서 변수까지 보여줌
#
#   - backtrace 옵션 : 스택 나열 (False 로 지정하면 마지막 스택만 나옴)
#   - diagnose 옵션  : 변수 나열 (False 로 지정하면 변수 안나옴)
#
#   - 권장 방법 : backtrace(True), diagnose(False)
#   - Apify 로그는 backtrace(False), diagnose(False) 로 되어있음
# --------------------------------------------------------------------------------------------------------------
def init_log(_file=None, hdr='edk'):
    logger.remove()
    if _file:
        app_dir = f'{os.path.dirname(os.path.abspath(_file))}'
        log_dir = os.path.join(app_dir, SX_LOGS)
        log_file = f'{log_dir}/{hdr}_{{time:YYYY-MM-DD}}.log'
        logger.add(
            sink=log_file,
            format='{time:YYYY-MM-DD HH:mm:ss.SSS} | <level>{level: <8}</level> | {message}',
            level='TRACE',
            backtrace=True,     # logger.exception() 에만 해당
            diagnose=False,     # logger.exception() 에만 해당
            rotation='00:00',
            retention='7 days'
        )
    # 화면은 backtrace(False), diagnose(False), 로그파일은 backtrace(True), diagnose(False)
    logger.add(
        sink=sys.stdout,
        format='{time:YYYY-MM-DD HH:mm:ss.SSS} | <level>{level: <8}</level> | {message}',
        level='TRACE',
        colorize=True,
        backtrace=False,
        diagnose=False
    )
    logger.level('DEBUG', color='<dim>')
    logger.level('INFO', color='<cyan>')

# --------------------------------------------------------------------------------------------------------------
# import시 공개할 심볼(함수, 클래스, 변수등)을 제한
# 오직 from module import * 에만 영향을 미침
# --------------------------------------------------------------------------------------------------------------
# __all__ = ['logger', 'init_log']


import os
import re
from datetime import datetime
from typing import List
from .constants import *
from .query import fix_query

# --------------------------------------------------------------------------------------------------------------
# 계정       install, (storage, domain), scraper
# musicow / clz    - coin             - news   - catalog - xxx
# musicow / clz    - coin             - news   - config
# --------------------------------------------------------------------------------------------------------------
class ApifyPathManager:
    def __init__(self, _storage, _name, _query, install=SX_STORAGE_BASE):
        self.name = _name
        self.fixname = self._gend(_query)
        self.INSTALL = install
        self.STORAGE = _storage         # domain
        self.query = _query
        self._update()

    def _update(self):
        self.PATH = f'{self.INSTALL}-{self.STORAGE}'
        self.NEWS_PATH = f'{self.PATH}-{AF_SCRAPER_NEWS}'
        self.CONFIG_PATH = f'{self.NEWS_PATH}-{AF_CONFIG}'
        self.CATALOG_PATH = f'{self.NEWS_PATH}-{AF_CATALOG}'
        self.BRONZE_PATH = f'{self.NEWS_PATH}-{AF_BRONZE}'
        self.SILVER_PATH = f'{self.NEWS_PATH}-{AF_SILVER}'
        self.GOLD01_PATH = f'{self.NEWS_PATH}-{AF_GOLD_S01}'
        self.GOLD02_PATH = f'{self.NEWS_PATH}-{AF_GOLD_S02}'
        self.MART01_PATH = f'{self.NEWS_PATH}-{AF_MART_S01}'

    def getname(self) -> str:
        return self.name
    def getfixname(self) -> str:
        return self.fixname
    def getquery(self) -> str:
        return self.query

    def _gend(self, _query: str) -> str:
        return f'{AF_QRY_HDR}{fix_query(self.name, _query)}'

    def base_catalog(self) -> str:
        return self.CATALOG_PATH
    def base_bronze(self) -> str:
        return self.BRONZE_PATH
    def base_silver(self) -> str:
        return self.SILVER_PATH
    def base_gold01(self) -> str:
        return self.GOLD01_PATH
    def base_gold02(self) -> str:
        return self.GOLD02_PATH
    def base_mart01(self) -> str:
        return self.MART01_PATH

    def kv_config(self) -> str:
        return f'{self.CONFIG_PATH}'
    def kv_catalog(self) -> str:
        return f'{self.CATALOG_PATH}-{self.fixname}'
    def kv_bronze(self) -> str:
        return f'{self.BRONZE_PATH}-{self.fixname}'
    def kv_silver(self) -> str:
        return f'{self.SILVER_PATH}-{self.fixname}'
    def kv_gold01(self) -> str:
        return f'{self.GOLD01_PATH}-{self.fixname}'
    def kv_gold02(self) -> str:
        return f'{self.GOLD02_PATH}-{self.fixname}'
    def kv_mart01(self) -> str:
        return f'{self.MART01_PATH}-{self.fixname}'

    def query_catalog(self) -> str:
        return f'{self.base_catalog()}-{self.fixname}'
    def query_catalog_index(self) -> str:
        return f'{self.base_catalog()}-{self.fixname}-{AF_INDEX}'
    def query_catalog_yearmonth(self, _date) -> str:
        if isinstance(_date, str):
            _date = datetime.strptime(_date, '%Y-%m')
        return f'{self.base_catalog()}-{self.fixname}-{_date.strftime("%Y-%m")}'
    #
    def query_bronze_index(self) -> str:
        return f'{self.base_bronze()}-{self.fixname}-{AF_INDEX}'
    def query_bronze_yearmonth(self, _date) -> str:
        if isinstance(_date, str):
            _date = datetime.strptime(_date, '%Y-%m')
        return f'{self.base_bronze()}-{self.fixname}-{_date.strftime("%Y-%m")}'
    #
    def query_silver_index(self) -> str:
        return f'{self.base_silver()}-{self.fixname}-{AF_INDEX}'
    def query_silver_yearmonth(self, _date) -> str:
        if isinstance(_date, str):
            _date = datetime.strptime(_date, '%Y-%m')
        return f'{self.base_silver()}-{self.fixname}-{_date.strftime("%Y-%m")}'
    #
    def query_gold01_index(self) -> str:
        return f'{self.base_gold01()}-{self.fixname}-{AF_INDEX}'
    def query_gold01_yearmonth(self, _date) -> str:
        if isinstance(_date, str):
            _date = datetime.strptime(_date, '%Y-%m')
        return f'{self.base_gold01()}-{self.fixname}-{_date.strftime("%Y-%m")}'
    #
    def query_gold02_index(self) -> str:
        return f'{self.base_gold02()}-{self.fixname}-{AF_INDEX}'
    def query_gold02_yearmonth(self, _date) -> str:
        if isinstance(_date, str):
            _date = datetime.strptime(_date, '%Y-%m')
        return f'{self.base_gold02()}-{self.fixname}-{_date.strftime("%Y-%m")}'
    #
    def query_mart01_index(self) -> str:
        return f'{self.base_mart01()}-{self.fixname}-{AF_INDEX}'
    def query_mart01_yearmonth(self, _date) -> str:
        if isinstance(_date, str):
            _date = datetime.strptime(_date, '%Y-%m')
        return f'{self.base_mart01()}-{self.fixname}-{_date.strftime("%Y-%m")}'

    # raw_04_00001_d83ca7c4e62e6241e8f1d825cf87802e1df975f1f3ef6866c44fe026d8e91204
    def bronze_data_name(self, _period_start, _num, _fingerprint) -> str:
        return f'{SX_RAW_HDR}{_period_start[-2:]}_{_num:05}_{_fingerprint}'

    def silver_data_name(self, _raw: str) -> str:
        base_name_with_ext = _raw[len(SX_RAW_HDR):]
        base_name, _ = os.path.splitext(base_name_with_ext)
        return f'{SX_PUR_HDR}{base_name}'

    def gold_step01_data_name(self, _pur: str) -> str:
        base_name_with_ext = _pur[len(SX_PUR_HDR):]
        base_name, _ = os.path.splitext(base_name_with_ext)
        return f'{SX_S01_HDR}{base_name}'

    def gold_step02_data_name(self, _s01: str) -> str:
        base_name_with_ext = _s01[len(SX_S01_HDR):]
        base_name, _ = os.path.splitext(base_name_with_ext)
        return f'{SX_S02_HDR}{base_name}'

    def mart_step01_data_name(self) -> str:
        return f'{SX_M01_HDR}stat01'

# --------------------------------------------------------------------------------------------------------------
# todo : 업데이트
class FilePathManager:
    def __init__(self, _name, _query, _storage, install=SX_STORAGE_BASE):
        self.name = _name
        self.fixname = self._gend(_query)
        self.query = _query
        # self.fixname = 'q6b46da7dddaa4417'
        self.STORAGE = _storage
        self.INSTALL = install
        self._update()

    def _update(self):
        self.PATH = os.path.join(self.INSTALL, self.STORAGE)
        self.CATALOG_PATH = os.path.join(self.PATH, SX_CATALOG)
        self.CATALOG_NEWS = os.path.join(self.CATALOG_PATH, SX_SCRAPER_NEWS)
        self.BRONZE_PATH = os.path.join(self.PATH, SX_BRONZE)
        self.BRONZE_RAW_PATH = os.path.join(self.BRONZE_PATH, SX_BRONZE_RAW)
        self.BRONZE_RAW_NEWS = os.path.join(self.BRONZE_RAW_PATH, SX_SCRAPER_NEWS)
        self.BRONZE_FILTER_PATH = os.path.join(self.BRONZE_PATH, SX_BRONZE_FILTER)
        self.BRONZE_FILTER_NEWS = os.path.join(self.BRONZE_FILTER_PATH, SX_SCRAPER_NEWS)
        self.SILVER_PATH = os.path.join(self.PATH, SX_SILVER)
        self.SILVER_NEWS = os.path.join(self.SILVER_PATH, SX_SCRAPER_NEWS)
        self.GOLD_PATH = os.path.join(self.PATH, SX_GOLD)
        self.GOLD_STEP01_PATH = os.path.join(self.GOLD_PATH, SX_GOLD_STEP01)
        self.GOLD_STEP01_NEWS = os.path.join(self.GOLD_STEP01_PATH, SX_SCRAPER_NEWS)
        self.GOLD_STEP02_PATH = os.path.join(self.GOLD_PATH, SX_GOLD_STEP02)
        self.GOLD_STEP02_NEWS = os.path.join(self.GOLD_STEP02_PATH, SX_SCRAPER_NEWS)

    def getname(self) -> str:
        return self.name
    def getfixname(self) -> str:
        return self.fixname
    def getquery(self) -> str:
        return self.query

    def _gend(self, _query: str) -> str:
        return f'{SX_QRY_HDR}{fix_query(self.name, _query)}'

    def storage_path(self, _path):
        marker = f'{self.STORAGE}/'
        index = _path.find(marker)
        if index != -1:
            return _path[index:]
        else:
            return 'not founded'

    # ----------------------------------------------------------------------------------------
    def catalog_query_dir(self) -> str:
        return os.path.join(self.CATALOG_NEWS, self.fixname)

    def catalog_tmp_dir(self, _date: datetime) -> str:
        return os.path.join(self.catalog_query_dir(), f'_{_date.strftime("%Y_%m")}')

    def catalog_tmp_name(self, _date: datetime) -> str:
        tmp_path = self.catalog_tmp_dir(_date)
        # return os.path.join(tmp_path, f'{SX_TMP_HDR}{SX_IDX_HDR}{_date.strftime("%Y_%m_%d")}{SX_EXT_YAML}')
        return os.path.join(tmp_path, f'{SX_TMP_HDR}{SX_IDX_HDR}{_date.strftime("%Y_%m_%d")}{SX_EXT_JSON}')

    def catalog_data_name(self, _date: datetime) -> str:
        # return os.path.join(self.catalog_query_dir(), f'{SX_IDX_HDR}{_date.strftime("%Y_%m")}{SX_EXT_YAML}')
        return os.path.join(self.catalog_query_dir(), f'{SX_IDX_HDR}{_date.strftime("%Y_%m")}{SX_EXT_JSON}')

    def catalog_data_list(self):
        # files = sorted(
        #     [f for f in os.listdir(self.catalog_query_dir()) if f.startswith(SX_IDX_HDR) and f.endswith(SX_EXT_YAML)],
        #     key=lambda x: datetime.strptime('_'.join(x.replace(SX_EXT_YAML, '').split('_')[2:4]), '%Y_%m')
        # )
        files = sorted(
            [f for f in os.listdir(self.catalog_query_dir()) if f.startswith(SX_IDX_HDR) and f.endswith(SX_EXT_JSON)],
            key=lambda x: datetime.strptime('_'.join(x.replace(SX_EXT_JSON, '').split('_')[2:4]), '%Y_%m')
        )
        return files

    # ----------------------------------------------------------------------------------------
    def bronze_query_raw_dir(self) -> str:
        return os.path.join(self.BRONZE_RAW_NEWS, self.fixname)

    def bronze_raw_year_dir(self, _indexfn: str) -> str:
        return os.path.join(self.bronze_query_raw_dir(), f'{SX_KEY_YEAR}{_indexfn.split("_")[2]}')
        
    def bronze_raw_data_dir(self, _indexfn: str) -> str:
        # return os.path.join(self.bronze_raw_year_dir(_indexfn), f'{SX_KEY_MONTH}{_indexfn.split("_")[3].replace(SX_EXT_YAML, "")}')
        return os.path.join(self.bronze_raw_year_dir(_indexfn), f'{SX_KEY_MONTH}{_indexfn.split("_")[3].replace(SX_EXT_JSON, "")}')

    # def bronze_raw_data_tmp_dir(self, _indexfn: str) -> str:
    #     return os.path.join(self.bronze_raw_year_dir(_indexfn), f'{SX_TMP_HDR}{_indexfn.split("_")[3].replace(SX_EXT_YAML, "")}')

    def bronze_raw_data_name(self, _indexfn: str, _period_start, _num, _fingerprint) -> str:
        return os.path.join(self.bronze_raw_data_dir(_indexfn), f'{SX_RAW_HDR}{_period_start[-2:]}_{_num:05}_{_fingerprint}{SX_EXT_JSON}')

    # def bronze_raw_data_tmp_name(self, _indexfn: str, _period_start, _num, _fingerprint) -> str:
    #     return os.path.join(self.bronze_raw_data_tmp_dir(_indexfn), f'{SX_RAW_HDR}{_period_start[-2:]}_{_num:05}_{_fingerprint}{SX_EXT_HTML}')

    # # ----------------------------------------------------------------------------------------
    def bronze_query_filter_dir(self) -> str:
        return os.path.join(self.BRONZE_FILTER_NEWS, self.fixname)

    def bronze_filter_year_dir(self, _indexfn: str) -> str:
        return os.path.join(self.bronze_query_filter_dir(), f'{SX_KEY_YEAR}{_indexfn.split("_")[2]}')

    def bronze_filter_data_dir(self, _indexfn: str) -> str:
        # return os.path.join(self.bronze_filter_year_dir(_indexfn), f'{SX_KEY_MONTH}{_indexfn.split("_")[3].replace(SX_EXT_YAML, "")}')
        return os.path.join(self.bronze_filter_year_dir(_indexfn), f'{SX_KEY_MONTH}{_indexfn.split("_")[3].replace(SX_EXT_JSON, "")}')

    # # def bronze_filter_data_tmp_dir(self, _indexfn: str) -> str:
    # #     return os.path.join(self.bronze_filter_year_dir(_indexfn), f'{SX_TMP_HDR}{_indexfn.split("_")[3].replace(SX_EXT_YAML, "")}')

    # def bronze_filter_data_name(self, _indexfn: str, _period_start, _num, _fingerprint) -> str:
    #     return os.path.join(self.bronze_filter_data_dir(_indexfn), f'{SX_FLT_HDR}{_period_start[-2:]}_{_num:05}_{_fingerprint}{SX_EXT_HTML}')

    # # def bronze_filter_data_tmp_name(self, _indexfn: str, _period_start, _num, _fingerprint) -> str:
    # #     return os.path.join(self.bronze_filter_data_tmp_dir(_indexfn), f'{SX_FLT_HDR}{_period_start[-2:]}_{_num:05}_{_fingerprint}{SX_EXT_HTML}')

    # ----------------------------------------------------------------------------------------
    def bronze_raw_year_dir_list(self) -> List:
        raw_dir = self.bronze_query_raw_dir()
        folders = sorted(
            [f.split('=')[1] for f in os.listdir(raw_dir) if os.path.isdir(os.path.join(raw_dir, f)) and re.match(rf'{SX_KEY_YEAR}\d{{4}}', f)],
            key=lambda x: datetime.strptime(x, '%Y')
        )
        # ['2022', '2023', '2024']
        return folders
   
    # '2024'
    def bronze_raw_data_dir_list(self, _year: str) -> List:
        data_dir = os.path.join(self.bronze_query_raw_dir(), f'{SX_KEY_YEAR}{_year}')
        folders = sorted(
            [f.split('=')[1] for f in os.listdir(data_dir) if os.path.isdir(os.path.join(data_dir, f)) and re.match(rf'{SX_KEY_MONTH}\d{{2}}', f)],
            key=lambda x: int(x)
        )
        # ['09', '10']
        return folders

    # raw_03_00003_126d6bd606088d7e9049789ab635a18de03322dffe11a2c3a63a76c5345eac2d.html
    def bronze_raw_data_list(self, _year: str, _month: str):
        data_dir = os.path.join(self.bronze_query_raw_dir(), f'{SX_KEY_YEAR}{_year}', f'{SX_KEY_MONTH}{_month}')
        files = sorted(
            [f for f in os.listdir(data_dir) if f.startswith(SX_RAW_HDR) and f.endswith(SX_EXT_JSON)],
            key=lambda x: (int(x.split('_')[1]), int(x.split('_')[2]))
        )
        return files

    # ----------------------------------------------------------------------------------------
    def silver_query_dir(self) -> str:
        return os.path.join(self.SILVER_NEWS, self.fixname)

    # /Users/eastjoo/works/scraper/x4t/03.silver/news/query_서태지/2024
    def silver_year_dir(self, _year: str) -> str:
        return os.path.join(self.silver_query_dir(), f'{SX_KEY_YEAR}{_year}')

    # /Users/eastjoo/works/scraper/x4t/03.silver/news/query_서태지/2024/07
    def silver_data_dir(self, _year: str, _month: str) -> str:
        return os.path.join(self.silver_year_dir(_year), f'{SX_KEY_MONTH}{_month}')

    def silver_data_name(self, _year: str, _month: str, _raw: str) -> str:
        base_name_with_ext = _raw[len('raw_'):]
        base_name, _ = os.path.splitext(base_name_with_ext)
        return os.path.join(self.silver_data_dir(_year, _month), f'{SX_PUR_HDR}{base_name}{SX_EXT_JSON}')

    # def silver_data_tmp_dir(self, _indexfn: str) -> str:
    #     return os.path.join(self.silver_year_dir(_indexfn), f'{SX_TMP_HDR}{_indexfn.split("_")[3].replace(SX_EXT_YAML, "")}')

    # def silver_data_tmp_name(self, _indexfn: str, _period_start, _num, _fingerprint) -> str:
    #     return os.path.join(self.silver_data_tmp_dir(_indexfn), f'{SX_PUR_HDR}{_period_start[-2:]}_{_num:05}_{_fingerprint}{SX_EXT_JSON}')

    def silver_year_dir_list(self) -> List:
        raw_dir = self.silver_query_dir()
        folders = sorted(
            [f.split('=')[1] for f in os.listdir(raw_dir) if os.path.isdir(os.path.join(raw_dir, f)) and re.match(rf'{SX_KEY_YEAR}\d{{4}}', f)],
            key=lambda x: datetime.strptime(x, '%Y')
        )
        # ['2022', '2023', '2024']
        return folders

    # '2024'
    def silver_data_dir_list(self, _year: str) -> List:
        data_dir = os.path.join(self.silver_query_dir(), f'{SX_KEY_YEAR}{_year}')
        folders = sorted(
            [f.split('=')[1] for f in os.listdir(data_dir) if os.path.isdir(os.path.join(data_dir, f)) and re.match(rf'{SX_KEY_MONTH}\d{{2}}', f)],
            key=lambda x: int(x)  # 추출한 숫자 부분을 정렬
        )
        # ['09', '10']
        return folders

    # pur_03_00003_126d6bd606088d7e9049789ab635a18de03322dffe11a2c3a63a76c5345eac2d.json
    def silver_data_list(self, _year: str, _month: str):
        data_dir = os.path.join(self.silver_query_dir(), f'{SX_KEY_YEAR}{_year}', f'{SX_KEY_MONTH}{_month}')
        files = sorted(
            [f for f in os.listdir(data_dir) if f.startswith(SX_PUR_HDR) and f.endswith(SX_EXT_JSON)],
            key=lambda x: (int(x.split('_')[1]), int(x.split('_')[2]))
        )
        return files

    # ----------------------------------------------------------------------------------------
    def gold_query_step01_dir(self) -> str:
        return os.path.join(self.GOLD_STEP01_NEWS, self.fixname)

    def gold_step01_year_dir(self, _year: str) -> str:
        return os.path.join(self.gold_query_step01_dir(), f'{SX_KEY_YEAR}{_year}')

    def gold_step01_data_dir(self, _year: str, _month: str) -> str:
        return os.path.join(self.gold_step01_year_dir(_year), f'{SX_KEY_MONTH}{_month}')

    def gold_step01_data_name(self, _year: str, _month: str, _raw: str) -> str:
        base_name_with_ext = _raw[len('pur_'):]
        base_name, _ = os.path.splitext(base_name_with_ext)
        return os.path.join(self.gold_step01_data_dir(_year, _month), f'{SX_S01_HDR}{base_name}{SX_EXT_JSON}')

    def gold_step01_year_dir_list(self) -> List:
        raw_dir = self.gold_query_step01_dir()
        folders = sorted(
            [f.split('=')[1] for f in os.listdir(raw_dir) if os.path.isdir(os.path.join(raw_dir, f)) and re.match(rf'{SX_KEY_YEAR}\d{{4}}', f)],
            key=lambda x: datetime.strptime(x, '%Y')
        )
        # ['2022', '2023', '2024']
        return folders

    # '2024'
    def gold_step01_data_dir_list(self, _year: str) -> List:
        data_dir = os.path.join(self.gold_query_step01_dir(), f'{SX_KEY_YEAR}{_year}')
        folders = sorted(
            [f.split('=')[1] for f in os.listdir(data_dir) if os.path.isdir(os.path.join(data_dir, f)) and re.match(rf'{SX_KEY_MONTH}\d{{2}}', f)],
            key=lambda x: int(x)  # 추출한 숫자 부분을 정렬
        )
        # ['09', '10']
        return folders

    # s01_03_00003_126d6bd606088d7e9049789ab635a18de03322dffe11a2c3a63a76c5345eac2d.json
    def gold_step01_data_list(self, _year: str, _month: str):
        data_dir = os.path.join(self.gold_query_step01_dir(), f'{SX_KEY_YEAR}{_year}', f'{SX_KEY_MONTH}{_month}')
        files = sorted(
            [f for f in os.listdir(data_dir) if f.startswith(SX_S01_HDR) and f.endswith(SX_EXT_JSON)],
            key=lambda x: (int(x.split('_')[1]), int(x.split('_')[2]))
        )
        return files

    # ----------------------------------------------------------------------------------------
    def gold_query_step02_dir(self) -> str:
        return os.path.join(self.GOLD_STEP02_NEWS, self.fixname)

    def gold_step02_year_dir(self, _year: str) -> str:
        return os.path.join(self.gold_query_step02_dir(), f'{SX_KEY_YEAR}{_year}')

    def gold_step02_data_dir(self, _year: str, _month: str) -> str:
        return os.path.join(self.gold_step02_year_dir(_year), f'{SX_KEY_MONTH}{_month}')

    def gold_step02_data_name(self, _year: str, _month: str, _raw: str) -> str:
        base_name_with_ext = _raw[len(SX_S01_HDR):]
        base_name, _ = os.path.splitext(base_name_with_ext)
        return os.path.join(self.gold_step02_data_dir(_year, _month), f'{SX_S02_HDR}{base_name}{SX_EXT_JSON}')

    def gold_step02_data_list(self, _year: str, _month: str):
        data_dir = os.path.join(self.gold_query_step02_dir(), f'{SX_KEY_YEAR}{_year}', f'{SX_KEY_MONTH}{_month}')
        files = sorted(
            [f for f in os.listdir(data_dir) if f.startswith(SX_S02_HDR) and f.endswith(SX_EXT_JSON)],
            key=lambda x: (int(x.split('_')[1]), int(x.split('_')[2]))
        )
        return files

# eof


import random

# -----------------------------------------------------------------------------------------------------------------
# (*) 미국 타임존
#   - Eastern Time (ET) 
#     timezone = 'America/New_York'
#   - Central Time (CT) 
#     timezone = 'America/Chicago'
#   - Mountain Time (MT) 
#     timezone = 'America/Denver'
#   - Pacific Time (PT) 
#     timezone = 'America/Los_Angeles'

# -----------------------------------------------------------------------------------------------------------------
# viewport : 브라우저 렌더링 영역의 크기를 설정. 사용자 환경 위장 (해상도 조작)
# window.outerWidth/Height (브라우저 전체) : 브라우저 실제 창 크기 조작 (headless 봇 탐지 회피)
def get_random_viewport():
    viewports = [
        {'width': 1920, 'height': 1080},    # Full HD
        {'width': 1600, 'height': 900},
        {'width': 1366, 'height': 768},
        {'width': 1440, 'height': 900},
        {'width': 1280, 'height': 800},
        {'width': 1536, 'height': 864},     # Laptop
        {'width': 1280, 'height': 720},     # 720p : 데스크톱 기준 해상도(1280x720 이상)를 설정하는 것이 안전
        #
        #
        # {'width': 1024, 'height': 1366}   # iPad Pro
        # {'width': 810, 'height': 1080},   # iPad Mini
        # {'width': 414, 'height': 896},    # iPhone XR
        # {'width': 390, 'height': 844},    # iPhone 14 Pro Max
        # {'width': 375, 'height': 667},    # iPhone SE
    ]
    outsize = random.choice(viewports)
    insize = {'width': outsize['width'], 'height': outsize['height'] - 120}
    return outsize, insize

# -----------------------------------------------------------------------------------------------------------------
# def get_random_viewport():
#     desktop_viewports = [
#         {'width': 1920, 'height': 1080},
#         {'width': 1600, 'height': 900},
#         {'width': 1280, 'height': 800}
#     ]
#     mobile_viewports = [
#         {'width': 390, 'height': 844},
#         {'width': 414, 'height': 896}
#     ]
#     # 데스크톱 70%, 모바일 30%
#     return random.choices(desktop_viewports + mobile_viewports, weights=[7, 7, 7, 3, 3])[0]

# -----------------------------------------------------------------------------------------------------------------
# 이게 다양해지면, 200 ok 가 나와도, html classid 가 바뀌어 파싱이 오류 (모바일 해상도는 삭제함)
user_agents = [
    # Windows - Chrome
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
    # Windows - Firefox
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:118.0) Gecko/20100101 Firefox/118.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:117.0) Gecko/20100101 Firefox/117.0",
    # Windows - Edge
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36 Edg/118.0.2088.61",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36 Edg/119.0.2151.58",
    # macOS - Chrome
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    # macOS - Safari
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.0 Safari/605.1.15",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 12_0_1) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.3 Safari/605.1.15",
    # macOS - Firefox
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7; rv:118.0) Gecko/20100101 Firefox/118.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_6) Gecko/20100101 Firefox/117.0",
    # Linux - Chrome
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    # Linux - Firefox
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:117.0) Gecko/20100101 Firefox/117.0",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Safari/537.36",
    #
    #
    # (*) iPhone - Safari
    # "Mozilla/5.0 (iPhone; CPU iPhone OS 16_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.0 Mobile/15E148 Safari/604.1",
    # "Mozilla/5.0 (iPhone; CPU iPhone OS 15_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.6 Mobile/15E148 Safari/604.1",
    # (*) iPad - Safari
    # "Mozilla/5.0 (iPad; CPU OS 16_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.0 Mobile/15E148 Safari/604.1",
    # (*) Android - Chrome
    # "Mozilla/5.0 (Linux; Android 12; SM-G991B) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Mobile Safari/537.36",
    # "Mozilla/5.0 (Linux; Android 11; Pixel 6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Mobile Safari/537.36",
    # (*) Android - Firefox
    # "Mozilla/5.0 (Android 11; Mobile; rv:117.0) Gecko/117.0 Firefox/117.0",
    # (*) Android - Edge
    # "Mozilla/5.0 (Linux; Android 12; Pixel 6 Pro) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Mobile Safari/537.36 EdgA/119.0.0.0",
    # (*) Other
    # "Mozilla/5.0 (X11; CrOS x86_64 14541.0.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36"
]

# -----------------------------------------------------------------------------------------------------------------
# 리퍼러가 요청마다 바뀌면 오히려 탐지될 수 있습니다.
# referers = [
#     "https://www.google.com",
#     "https://www.bing.com",
#     "https://www.yahoo.com",
#     "https://duckduckgo.com"
# ]
# referers = [
#     "https://news.google.com",
#     "https://www.nytimes.com",
#     "https://www.reddit.com",
#     "https://www.wikipedia.org"
# ]
#
# "Accept-Language": "ko-KR,ko;q=0.9,en-US;q=0.7,en;q=0.6",
#
# headers = {
#     "User-Agent": random.choice(user_agents),
#     "Referer": random.choice(referers),  # 리퍼러 랜덤 설정
#     "Accept-Language": "en-US,en;q=0.9",
#     "Connection": "keep-alive"
# }

# -----------------------------------------------------------------------------------------------------------------
# proxies = {
#     'http': 'http://username:password@proxy_server:port',
#     'https': 'https://username:password@proxy_server:port'
# }



import random

# -----------------------------------------------------------------------------------------------------------------
async def strategy_webdriver(_context):
    # 브라우저 지문 조작 (navigator.webdriver 제거)
    # 구글은 navigator.webdriver 값이 true이면 봇으로 간주
    await _context.add_init_script("""
        Object.defineProperty(navigator, 'webdriver', {
            get: () => undefined
        })
    """)

# -----------------------------------------------------------------------------------------------------------------
async def strategy_chrome(_context):
    # window.chrome 속성 추가 (Chrome DevTools 환경 우회)
    # 실제 브라우저에는 window.chrome 객체가 존재하지만, 자동화된 환경에서는 없으므로 이를 추가함
    # - window.chrome : 크롬 개발자 도구(DevTools)에서 사용되는 API와 관련됩니다.
    # - window.navigator.chrome : 브라우저 및 사용자 환경 정보
    await _context.add_init_script("""
        window.chrome = {
            app: {},
            runtime: {},
            loadTimes: function() {},
            csi: function() {}
        };
        window.navigator.chrome = {
            app: {},
            runtime: {},
            loadTimes: function() {},
            csi: function() {}
        };
    """)

# -----------------------------------------------------------------------------------------------------------------
async def strategy_plugin(_context):
    # 플러그인 개수 설정 (빈 배열이면 자동화로 간주)
    await _context.add_init_script("""
        Object.defineProperty(navigator, 'plugins', {
            get: () => [1, 2, 3, 4, 5]
        });
    """)

# -----------------------------------------------------------------------------------------------------------------
async def strategy_lang(_context):
    # 언어 설정 (실제 브라우저와 일치)
    await _context.add_init_script("""
        Object.defineProperty(navigator, 'languages', {
            get: () => ['ko-KR', 'en-US']
        });
    """)

# -----------------------------------------------------------------------------------------------------------------
async def strategy_tz(_context):
    # 시간대 설정 (Intl API 조작)
    await _context.add_init_script("""
        const timeZone = 'Asia/Seoul';
        Object.defineProperty(Intl.DateTimeFormat.prototype, 'resolvedOptions', {
            value: function() {
                return {
                    timeZone: timeZone
                };
            }
        });
    """)

# -----------------------------------------------------------------------------------------------------------------
# WebGL 렌더러 및 벤더 스푸핑 
async def strategy_webgl(_context):
    renderers = [ "NVIDIA GeForce GTX 1650", "AMD Radeon RX 580", "Intel Iris Xe Graphics"]
    vendors = [ "NVIDIA Corporation", "Advanced Micro Devices, Inc.", "Intel Corporation"]
    selected_renderer = random.choice(renderers)
    selected_vendor = random.choice(vendors)
    # f-string에서 중괄호 두 번 사용해 JavaScript 코드 삽입
    await _context.add_init_script(f"""
        const getParameter = WebGLRenderingContext.prototype.getParameter;
        WebGLRenderingContext.prototype.getParameter = function(parameter) {{
            if (parameter === 37445) {{
                return "{selected_renderer}";  // 렌더러 스푸핑
            }}
            if (parameter === 37446) {{
                return "{selected_vendor}";  // 벤더 스푸핑
            }}
            return getParameter(parameter);
        }};
    """)

# -----------------------------------------------------------------------------------------------------------------
async def strategy_canvas(_context):
    # Canvas Fingerprinting 조작
    # 구글은 HTMLCanvasElement를 사용해 브라우저에서 렌더링된 이미지를 기반으로 지문을 생성해 감지
    # 일부 픽셀 값을 수정해 캔버스 지문이 일치하지 않도록 설정
    # if (type === '2d') {  이게 오리지널...type 은 추가함
    await _context.add_init_script("""
        const getContext = HTMLCanvasElement.prototype.getContext;
        HTMLCanvasElement.prototype.getContext = function(type, attributes) {
            if (type === '2d' || type === 'webgl') {
                const context = getContext.call(this, type, attributes);
                const originalGetImageData = context.getImageData;
                context.getImageData = function(x, y, width, height) {
                    const imageData = originalGetImageData.call(this, x, y, width, height);
                    imageData.data[0] = imageData.data[0] ^ 0xff;  // 첫 번째 픽셀 수정
                    return imageData;
                };
            }
            return getContext.call(this, type, attributes);
        };
    """)

# -----------------------------------------------------------------------------------------------------------------
async def strategy_outer(_context, _outer_size):
    # 실제 브라우저처럼 window.outerHeight와 window.outerWidth를 설정해 헤드리스 환경을 감춤
    height = _outer_size['height']
    width = _outer_size['width']
    await _context.add_init_script(f"""
        window.outerHeight = {height};
        window.outerWidth = {width};
    """)

# -----------------------------------------------------------------------------------------------------------------
async def strategy_block(_context, _outer_size):
    await strategy_webdriver(_context)
    await strategy_chrome(_context)
    await strategy_plugin(_context)
    await strategy_lang(_context)
    await strategy_webgl(_context)
    await strategy_canvas(_context)
    await strategy_tz(_context)
    await strategy_outer(_context, _outer_size)

# eof


import random
import time
from .constants import *
from .block import *

# -----------------------------------------------------------------------------------------------------------------
async def load_cookie(_task_name, _actor):
    kvstore = await _actor.open_key_value_store(name=SX_COOKIE_STORE)
    index = random.randint(0, len(user_agents) - 1)
    cookie_key = f'cookies_{index}'
    # {'cookies': {...}, 'timestamp': 1693698539.1234}
    stored_data = await kvstore.get_value(cookie_key)
    cookies_data = None
    if stored_data:
        stored_timestamp = stored_data.get('timestamp', 0)
        # 1시간(3600초) 이내라면 쿠키 재사용
        if time.time() - stored_timestamp < 3600:
            cookies_data = stored_data.get('cookies')
        else:
            # logger.debug(f'[t] task-{int(_task_name):04d} the cookie has expired as more than 1 hour has passed.')
            await kvstore.set_value(cookie_key, None)   # 기존 쿠키 삭제(또는 set_value(cookie_key, {}) 등)
    return index, cookies_data, kvstore

# -----------------------------------------------------------------------------------------------------------------
async def save_cookie(_context, _index, _kvstore):
    data_to_store = {'cookies': await _context.storage_state(), 'timestamp': time.time()}
    await _kvstore.set_value(f'cookies_{_index}', data_to_store)

# -----------------------------------------------------------------------------------------------------------------
async def set_internal_proxy(_actor, _env_ctx):
    # Apify 플랫폼에서 Actor가 실행될 때, Apify 시스템이 APIFY_PROXY_HOSTNAME, APIFY_PROXY_PORT, APIFY_PROXY_PASSWORD 환경 변수
    # 를 자동으로 주입.
    proxy_configuration = await _actor.create_proxy_configuration(actor_proxy_input=_env_ctx['proxy'])
    return proxy_configuration

# -----------------------------------------------------------------------------------------------------------------
async def set_outernal_proxy(_actor, _env_ctx):
    # mac 로컬에서 실행시에는 password 를 지정해 줘야 함. 즉 이경우에는 외부 프록시 사용하는 경우임.
    proxy_configuration = await _actor.create_proxy_configuration(
        groups=_env_ctx['proxy']['apifyProxyGroups'],
        country_code=_env_ctx['proxy']['apifyProxyCountry'],
        password=_env_ctx['proxy_password']
    )
    return proxy_configuration

# -----------------------------------------------------------------------------------------------------------------
async def set_proxy(_task_name, _config):
    # (*) 환경 변수를 사용하여 내부 프록시를 사용해야 함
    # 외부 프록시 : http://groups-RESIDENTIAL:apify_proxy_u...d@proxy.apify.com:8000
    #
    # 내부 프록시 사용 중
    # http://groups-RESIDENTIAL,country-US:*********@10.0.93.195:8011
    proxy_url = await _config.new_url()

    # {'server': '10.0.93.195:8011', 'username': 'groups-RESIDENTIAL,country-US', 'password': '*********'}
    proxy_ctx = {
        'server': proxy_url.split('@')[-1],                                 # proxy.apify.com:8000
        'username': proxy_url.split('@')[0].split('//')[-1].split(':')[0],  # groups-RESIDENTIAL,country-US, or 'auto'
        'password': proxy_url.split('@')[0].split(':')[-1]                  # apify_proxy_xxxxxxxxxx
    }
    password = proxy_ctx['password']
    masked_password = f"{password[:-30]}{'*' * 30}"  # 마지막 5자리만 남기고 마스킹
    logging_ctx = proxy_ctx.copy()
    logging_ctx['password'] = masked_password
    # logger.debug(f'[t] task-{int(_task_name):04d} using a proxy : {logging_ctx['server']} ({logging_ctx['username']})')
    return proxy_ctx

# eof


CODE_OK = 0
CODE_1001 = 1001    # internal error
CODE_2001 = 2001    # llm error
CODE_3001 = 3001    # llm warning

# --------------------------------------------------------------------------------------------------------------
class ClzError(Exception):
    def __init__(self, message="CLZ Error", code=CODE_1001):
        self.message = message
        self.code = code
        self.fmt = self._format_message()
        super().__init__(self.fmt)

    def _format_message(self):
        # ( 2001) CLZ Internal Error
        return f'({self.code:5d}) {self.message}'

    def __str__(self):
        return self.fmt
    
# --------------------------------------------------------------------------------------------------------------
class ClzInternalError(ClzError):
    def __init__(self, message='CLZ Internal Error', code=CODE_2001, input_value=None):
        self.input_value = input_value
        super().__init__(message, code)
        if input_value:
            self.fmt = f'{self.fmt} | {input_value}'

    def __str__(self):
        return self.fmt

# --------------------------------------------------------------------------------------------------------------
class ClzWarn(ClzError):
    def __init__(self, message="CLZ Warning", code=CODE_3001):
        super().__init__(message, code)

    def __str__(self):
        return self.fmt

# eof



import hashlib
import yaml
from apify import Actor
from .constants import *
from .utils import elapsed, prompt_template_yaml

# opt 종류
QUERY = 'query'
ENGINE = 'engine'
FIX = 'fix'
PROMPT = 'prompt'
KEYWORDS = 'keywords'
DESCRIPTION = 'description'

# --------------------------------------------------------------------------------------------------------------
def get(_domain, _name, _opt):
    if _domain == SX_DOMAIN_NEWS:
        query_domain = f'query_{SX_DOMAIN_NEWS}'
    elif _domain == SX_DOMAIN_COMPANY:
        query_domain = f'query_{SX_DOMAIN_COMPANY}'
    elif _domain == SX_DOMAIN_COIN:
        query_domain = f'query_{SX_DOMAIN_COIN}'
    else:
        return None

    datadict = prompt_template_yaml(query_domain)
    for entry in datadict:
        if entry.get('name') == _name:
            if _opt == FIX:
                return fix_query(_name.strip(), entry[QUERY].strip())
            return entry[_opt].strip()
    return None

# --------------------------------------------------------------------------------------------------------------
# domain 별로 이름은 유니크 해야 함
# 'Hello | World + 한글 - "Test"'
# --------------------------------------------------------------------------------------------------------------
def fix_query(_name, _query):
    combined_query = _name + _query
    hash_bytes = hashlib.sha256(combined_query.encode()).digest()
    out =  bytes(a ^ b ^ c ^ d for a, b, c, d in zip(hash_bytes[:8], hash_bytes[8:16], hash_bytes[16:24], hash_bytes[24:]))
    return out.hex()

# --------------------------------------------------------------------------------------------------------------
def get_config_path(_storage, install=SX_STORAGE_BASE):
    basic_path = f'{install}-{_storage}'
    news_path = f'{basic_path}-{AF_SCRAPER_NEWS}'
    return f'{news_path}-{AF_CONFIG}'

# --------------------------------------------------------------------------------------------------------------
async def getkv(_config, _domain, _name, _opt):
    if _domain == SX_DOMAIN_NEWS:
        query_domain = f'query_{SX_DOMAIN_NEWS}'
    elif _domain == SX_DOMAIN_COMPANY:
        query_domain = f'query_{SX_DOMAIN_COMPANY}'
    elif _domain == SX_DOMAIN_COIN:
        query_domain = f'query_{SX_DOMAIN_COIN}'
    else:
        return None

    kv_store = await Actor.open_key_value_store(name=_config)
    begintrx = await kv_store.get_value(f'{query_domain}{SX_EXT_YAML}') or None
    if not begintrx:
        return get(_domain, _name, _opt)
    datadict = yaml.load(begintrx, Loader=yaml.FullLoader)
    for entry in datadict:
        if entry.get('name') == _name:
            if _opt == FIX:
                return fix_query(_name.strip(), entry[QUERY].strip())
            return entry[_opt].strip()
    return None

# eof